package drawing;

// 図形描画ツールの選択状態を表す抽象クラス
abstract class DTState {
  public abstract void handleClick( DrawTool dt, int x1, int y1 );
  public abstract void handleHelp();
  protected final void changeDTState( DrawTool dt, DTState s ) {
    dt.changeState( s );
  }
}

// 図形描画ツール無選択状態を表す具体クラス(Singletonパターン使用)
final class NonSelected extends DTState {
  private static NonSelected _instance = null;
  private NonSelected() {}
  public static DTState instance() {
    if( _instance == null ) _instance = new NonSelected();
    return _instance;
  }
  public void handleClick( DrawTool dt, int x1, int y1 ) {
    if( /* クリック位置が矩形描画ツールアイコンか */ ) {
      changeDTState( dt, RectToolSelected.instance() );
    }        
    else if( /* クリック位置がテキスト描画ツールアイコンか */ ) {
      changeDTState( dt, TextToolSelected.instance() );
    }
  }
  public void handleHelp() {
    // 全般的な事項に関するヘルプ画面を表示する
  }
}

// 矩形描画ツール選択状態を表す具体クラス(Singletonパターン使用)
final class RectToolSelected extends DTState {
  private static RectToolSelected _instance = null;
  private RectToolSelected() {}
  public static DTState instance() {
    if( _instance == null ) _instance = new RectToolSelected();
      return _instance;
  }
  public void handleClick( DrawTool dt, int x1, int y1 ) {
    do {} while( /* マウスクリックが押されていないなら真 */ );
    if( /* クリック位置がドキュメント上の場合 */ ) {
      // マウスクリックの位置をx2とy2に格納し，点(x1,y1)
      // を矩形のひとつの角，その角と対角関係にある別の角
      // を点(x2, y2)として，矩形を描画する
    }
    else if( /* クリック位置がテキスト描画ツールアイコンか */ ) {
      changeDTState( dt, TextToolSelected.instance() );
    }
    else if( /* クリックで描画ツール選択が解除になるか */ ) {
      changeDTState( dt, NonSelected.instance() );
    }
  }
  public void handleHelp() {
    // 矩形描画ツールの使い方を説明するヘルプ画面を表示する
  }
}

// テキスト描画ツール選択状態を表す具体クラス(Singletonパターン使用)
final class TextToolSelected extends DTState {
  private static TextToolSelected _instance = null;
  private TextToolSelected() {}
  public static DTState instance() {
    if( _instance == null ) _instance = new TextToolSelected();
    return _instance;
  }
  public void handleClick( DrawTool dt, int x1, int y1 ) {
    // テキスト入力フィールドを生成する
    do { /* テキスト入力を行う */ }
      while( /* テキストフィールド外でのクリックではないなら真 */ );
    // テキスト入力処理を終了する
    if( /* クリック位置が矩形描画ツールアイコンか */ ) {
      changeDTState( dt, RectToolSelected.instance() );
    }        
    else if( /* クリックで描画ツール選択が解除になるか */ ) {
      changeDTState( dt, NonSelected.instance() );
    }
  }
  public void handleHelp() {
    // テキスト描画ツールの使い方を説明するヘルプ画面を表示する
  }
}

// 図形描画ツールを表すクラス
final class DrawTool {
  private DTState tState = NonSelected.instance();
  public final void doClick( int x1, int y1 ) {//マウスクリック処理を行う
    tState.handleClick( this, x1, y1 );
  }
  public final void doHelp() { // ヘルプ機能を呼び出す
    tState.handleHelp();
  }
  final void changeState( DTState s ) { tState = s; }
}