import java.util.*; // Java2 (JDK 1.2 or later)

interface Command extends Cloneable { // Commandインタフェイス
  Object clone() throws CloneNotSupportedException;
  void execute( boolean redo ); // redoには“再実行”時にtrueを渡す
  void unexecute();
}

class MenuItem { // Invokerに相当する
  private Command command;
  public MenuItem() { command = null; }
  public void setCommand( Command c ) { command = c; }
  public void invoke() {
    if( command != null ) command.execute( false );
  }
}

class CommandList { // 命令履歴リスト
  private Vector commands; int index;
  public CommandList() {
    commands = new Vector(); index = commands.size();
  }
  public void pushCommand( Command c ) {
    while( index != commands.size() )
      commands.removeElementAt( commands.size() - 1 );
    try {
      Command clone = (Command) c.clone();
      if( clone != null ) commands.add( clone );
      index = commands.size();
    }
    catch ( CloneNotSupportedException e ) { /* 略 */  }
  }
  public void undoAction() { // “取り消し”のアクション
    if( index == 0 ) {
      System.out.println( "can't undo" ); return;
    }
    ((Command)(commands.elementAt( --index ))).unexecute();
  }
  public void redoAction() { // “再実行”のアクション
    if( index == commands.size() ) {
      System.out.println( "can't redo" ); return;
    }
    ((Command)(commands.elementAt( index++ ))).execute( true );
  }
}

class TextDocument { // Receiverに相当する
  private StringBuffer state; // 文字列データ // 内部状態
  private int insertionpoint; // 挿入ポイント
  public CommandList clist; // 命令履歴リストをドキュメントに持たせる
  public TextDocument() {
    state = new StringBuffer( "012345" ); insertionpoint = 0;
    clist = new CommandList(); 
  }
  public StringBuffer getData() { return state; }
  public void setData( StringBuffer str ) { state = str; }
  public void setInsertionPoint( int i ) { insertionpoint = i; }
  public int getInsertionPoint() { return insertionpoint; }
  public void insertAction( StringBuffer s ) {
    state.insert( insertionpoint, s.toString() );
  }
  public void removeAction( int pos, int len ) {
    state.delete( pos, pos + len );
  }
}

class Application { // Clientに相当する
  private TextDocument document;
  private MenuItem insertMenuItem, removeLastCharMenuItem,
                   undoMenuItem, redoMenuItem;
  public Application() { // Applicationのコンストラクタ
    document = new TextDocument();
    insertMenuItem = new MenuItem();
    removeLastCharMenuItem = new MenuItem();
    undoMenuItem   = new MenuItem(); redoMenuItem = new MenuItem();
    insertMenuItem.setCommand( new InsertCommand( this ) );
    removeLastCharMenuItem.setCommand( new RemoveLastCharCommand( this ) );
    undoMenuItem.setCommand( new UndoCommand( this ) );
    redoMenuItem.setCommand( new RedoCommand( this ) );
  }
  // 現在のドキュメントを得るメソッド
  public TextDocument getCurrentDocument() { return document; }
  // ドキュメントの内容を表示するメソッド
  public void print() { System.out.println( document.getData() ); }
  
  public void test() { // テスト用メソッド
    print();
    document.setInsertionPoint( 2 ); insertMenuItem.invoke(); print();
    document.setInsertionPoint( 4 ); insertMenuItem.invoke(); print();
    removeLastCharMenuItem.invoke(); print();
    System.out.println( "------------" );
    undoMenuItem.invoke(); print(); undoMenuItem.invoke(); print();
    undoMenuItem.invoke(); print();
    System.out.println( "------------" );
    redoMenuItem.invoke(); print(); redoMenuItem.invoke(); print();
    System.out.println( "------------" );
    document.setInsertionPoint( 1 ); insertMenuItem.invoke(); print();
    undoMenuItem.invoke(); print(); redoMenuItem.invoke(); print();
    redoMenuItem.invoke(); print();
  }
}

// テキスト中に文字列"abc"を挿入する命令
class InsertCommand implements Command {
  private Application appl; private TextDocument document;
  private int insertionpoint; private StringBuffer inserted_string;
  public Object clone() throws CloneNotSupportedException {
    InsertCommand ic = (InsertCommand)super.clone();
    ic.inserted_string = inserted_string;
    return ic;
  }
  public InsertCommand( Application a ) { appl = a; }
  public void execute( boolean redo ) {
    StringBuffer s = new StringBuffer( "abc" );
    if( redo ) document.setInsertionPoint( insertionpoint );
    (document = appl.getCurrentDocument()).insertAction( s );
    insertionpoint = document.getInsertionPoint();
    inserted_string = s; 
    if( ! redo ) document.clist.pushCommand( this );
  }
  public void unexecute() {
    document.removeAction( insertionpoint, inserted_string.length() );
    document.setInsertionPoint( insertionpoint );
  }
}

// テキストの末尾1文字を削除する命令
class RemoveLastCharCommand implements Command {
  private Application appl; private TextDocument document;
  private char deleted_char;
  public  RemoveLastCharCommand( Application a ) { appl = a; }
  public Object clone() throws CloneNotSupportedException {
    RemoveLastCharCommand rlcc = (RemoveLastCharCommand)super.clone();
    return rlcc;
  }
  public void execute( boolean redo ) {
    document = appl.getCurrentDocument();
    StringBuffer sb = document.getData();
    deleted_char = sb.charAt( sb.length() - 1 );
    document.removeAction( sb.length() - 1, 1 );
    if( ! redo ) document.clist.pushCommand( this );
  }
  public void unexecute() {
    int insertionpoint = document.getInsertionPoint();
    document.setInsertionPoint( document.getData().length() );
    Character c = new Character( deleted_char );
    document.insertAction( new StringBuffer( c.toString() ) );
    document.setInsertionPoint( insertionpoint );
  }
}

// “取り消し”命令
class UndoCommand implements Command {
  private Application appl; private TextDocument document;
  public UndoCommand( Application a ) { appl = a; }
  public Object clone() throws CloneNotSupportedException { return null; }
  public void execute( boolean redo ) {
    document = appl.getCurrentDocument(); document.clist.undoAction();
  }
  public void unexecute() {}
}

// “再実行”命令
class RedoCommand implements Command {
  private Application appl; private TextDocument document;
  public RedoCommand( Application a ) { appl = a; }
  public Object clone() throws CloneNotSupportedException { return null; }
  public void execute( boolean redo ) {
      document = appl.getCurrentDocument(); document.clist.redoAction();
  }
  public void unexecute() {}
}

public class Command1 {
  public static void main( String args[] ) {
    Application appl = new Application(); appl.test();
  }
}